/********************************************************************************
    Copyright (C) 1999  Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "error.hh"
#include "tests/system1.hh"
#include "tests/memalloc.hh"
#include "tests/vidaccess.hh"
#include "tests/videodec.hh"
#include "ulib/init.hh"
#include "config.h"


#include "input/streamsrc_istr.hh"
#include "input/errors.hh"
#include "system/sysdec1.hh"
#include "system/resync.hh"
#include "system/sysdemux.hh"
#include "system/videoauconv.hh"
#include "system/syncer.hh"
#include "video12/vdecoder.hh"
#include "vpostproc/buffer.hh"
#include "vpostproc/pp_mblks.hh"
#include "vpostproc/pp_imgtype.hh"
#include "vpostproc/pp_qscale.hh"
#include "vpostproc/pp_mv.hh"
#include "vpostproc/pp_mvcol.hh"

#include "ulib/x11/imgwin.hh"
#include "ulib/x11/server.hh"
#include "ulib/graphics/fileio/rw_ppm.hh"
#include "ulib/graphics/fileio/write_yuv.hh"
#include "ulib/graphics/color/colorspace.hh"

#include "options.hh"

#include <iostream.h>
#include <fstream.h>

#include <stdio.h>
#include <stdlib.h>


#if 1
#include "ulib/arch/x86/CPUcapabilities.hh"
#endif


// stream sources

static StreamSource_IStream strsrc_istr;
static StreamSource* streamsrc;


// the system decoder

static SystemDecoder_MPEG1 system_dec;
static SystemResyncer      system_resync;
static SystemDemux         system_demux(system_resync);

// video decoding pipeline

static SystemDemuxExtractor video_extractor;
static VideoAUConverter     vidcon(video_extractor);
static VideoDecoder         video_dec(vidcon);
static VideoImageBuffer     video_buffer;

// audio decoding pipeline

static SystemDemuxExtractor audio_extractor;
static SystemDemuxExtractor ac3_extractor;
static int ac3stream = 0x80;

// video postprocessors

static DecodedPictureSink* lastoutput;
static VideoPostprocessor_MBBoundaries pp_mblks;
static VideoPostprocessor_ImageType    pp_imgtype;
static VideoPostprocessor_QScale       pp_qscale;
static VideoPostprocessor_MotionVector pp_mv;
static VideoPostprocessor_MVCol        pp_mvcol;


// synchronization

static Syncer_Realtime syncer_realtime;



void InitStreamSource(ifstream& istr)
{
  strsrc_istr.SetIStream(istr);
  streamsrc=&strsrc_istr;
}


void InitVideoDecoder1()
{
  system_dec.SetSource(*streamsrc);
  system_resync.SetSource(&system_dec);

  system_demux.EnableChannel(StreamIDBase_Video|0);

  video_extractor.SetSource(&system_demux);
  video_extractor.SetChannel(StreamIDBase_Video|0);


  //video_dec.SetSyncer(&syncer_realtime);

  lastoutput = &video_buffer;
}


void InitAudioDecoder()
{
  system_demux.EnableChannel(StreamIDBase_Audio|0);
  audio_extractor.SetSource(&system_demux);
  audio_extractor.SetChannel(StreamIDBase_Audio|0);
}

void InitAC3Decoder()
{
  system_demux.EnableChannel(0xBD);
  ac3_extractor.SetSource(&system_demux);
  ac3_extractor.SetChannel(0xBD);
}


void InsertPostprocessor(VideoPostprocessor& pp)
{
  pp.SetNext(lastoutput);
  lastoutput = &pp;
}


void InitVideoDecoder2()
{
  video_dec.SetBufferProvider(*lastoutput);
}


class Output_Class
{
public:
  virtual ~Output_Class() { }

  virtual void Show(Image_YUV<Pixel>& img) = 0;
};


class Output_Window : public Output_Class
{
public:
  Output_Window() : first(true) { }

  void Show(Image_YUV<Pixel>& img)
  {
    if (first)
      {
	first=false;

	ImageParam_YUV param;
	img.GetParam(param);

	imgwin.Create(param.width,param.height,
		      "DVDview (c) 1998-2000 Dirk Farin");
      }

    imgwin.Display(img);
  }

private:
  ImageWindow_Autorefresh_X11 imgwin;
  bool first;
};


class Output_PPM : public Output_Class
{
public:
  Output_PPM() : nr(1) { }

  void Show(Image_YUV<Pixel>& img)
  {
    char buffer[100];
    sprintf(buffer,"img%05d.ppm",nr);
    ofstream ostr(buffer);

    Image_RGB<Pixel> rgbimg;

    YUV2RGB_420(img,rgbimg);
    WriteImage_PPM6(rgbimg,ostr);

    nr++;
  }

private:
  int nr;
};


class Output_YUV : public Output_Class
{
public:
  Output_YUV()
  {
    ostr.open("sequence.yuv");
    writer.SetYUVStream(ostr);
  }

  void Show(Image_YUV<Pixel>& img)
  {
    writer.WriteImage(img);
  }

private:
  FileWriter_YUV1 writer;
  ofstream ostr;
};


static Output_Class* out_class;


void ViewMPEG(bool audio,bool ac3)
{
  int w,h;

  int nframes=0;
  try
    {
      while (1)
	{
	  while (audio && audio_extractor.PacketPending())
	    {
	      SysPacket_Packet* pck = audio_extractor.GetNextPacket();
	      fwrite(pck->data.AskContents(),pck->data.AskLength(),1,stdout);
	      delete pck;
	    }

	  while (ac3 && ac3_extractor.PacketPending())
	    {
	      SysPacket_Packet* pck = ac3_extractor.GetNextPacket();
	      if (pck->data.AskContents()[0] == ac3stream)
		{
		  fwrite(pck->data.AskContents()+4,
			 pck->data.AskLength()-4,1,stdout);
		}

	      delete pck;
	    }

	  if (video_buffer.BufferEmpty())
	    {
	      bool more = video_dec.DecodeAFrame();
	      if (!more && video_buffer.BufferEmpty())
		break;
	    }
	  else
	    {
	      DecodedImageData* dimg;
	      dimg = video_buffer.GetNextImage();

	      out_class->Show(dimg->m_image);

	      video_buffer.FreeImage(dimg);
	      nframes++;
	    }
	}
    }
  catch(...)
    {
      cout << "Unexpected end. Decoded up to frame: " << nframes << endl;
      throw;
    }

  cout << "Decoded " << nframes << " frames.\n";
}


static void usage()
{
  cerr <<
    " DVDview             05.Jun.2000 (c) Dirk Farin\n"
    "------------------------------------------------\n"
    "usage: dvdview <options> filename.mpg\n"
    "\n"
    "Options:\n"
    "  -B    Skip B-frames\n"
    "  -P    Skip P-frames (and B-frames)\n"
    "  -v #  Enable logging level (0 <= # <= 7)\n"
    "         0 - off\n"
    "         1 - sequence headers\n"
    "         2 - gop headers\n"
    "         3 - picture headers\n"
    "         4 - slice headers\n"
    "         5 - macroblocks\n"
    "         6 - DCT coefficients\n"
    "         7 - binary slice data\n"
    "  -F    Write PPM sequence rather than display stream contents.\n"
    "  -Y    Write single YUV file rather than display stream contents.\n"
#if 0
    "  -u    Write UYVY sequence rather than display stream contents.\n"
    "  -w    Warn when false MPEG-1 entries in MPEG-2 stream found.\n"
    "  -b    Use only one MV when predicting bidirectionally. This results\n"
    "        in reduced accuracy but increased performance.\n"
    "  -H    Do not use half-pel MV accuracy in B-Pictures.\n"
#endif
    "  -M    Show macroblock boundaries (twice for bigger marks).\n"
    "  -d      Show boundary as dark pixels.\n"
    "  -l      Show boundary as bright pixels.\n"    
    "  -Q    Show QScale of macroblocks.\n"
    "  -V    Show motionvectors of macroblocks (vectors).\n"
    "  -C    Show motionvectors of macroblocks (colors).\n"
    "  -p      Enable vectors in P-frames.\n"
    "  -b      Enable vectors in B-frames.\n"
    "  -c      Show colored motionvectors.\n"
    "  -f      Enable forward vectors.\n"
    "  -r      Enable backward vectors.\n"
    "  -O      hold mode\n"
    "  -A    Extract audio-stream to stdout. Pipe this into mpg123.\n"
    "  -3    Extract AC3-stream to stdout. Pipe this into ac3dec.\n"
    "  -a #    select AC3-stream (0-7) default is 0.\n"
#if 0
    "  -W    Set time to wait after each frame (in 1/100s).\n"
    "  -X    Disable MMX.\n"
    "  -S    Disable X11 shared memory extension.\n"
    "  -F    DisableSomething=true (development only).\n"
    "  -m    Mark special MBs (development only).\n"
#endif
    "  -h    Show this usage information\n"
    ;
}



int main(int argc,char** argv)
{
  InitULib(argc,argv);

  VideoDecoder_Options vdec_options;

  bool imgtypes=false;
  bool qscales=false;
  bool mvs=false,mvcol=false;
  bool audio=false,ac3=false;
  bool mvf=false,mvb=false,mvcolor=false;
  bool mvpf=false,mvbf=false;
  bool mv_holdmode=false;

#if 0 && ENABLE_MMX
  cout << "CPU capabilities detected: ";
  if (cpucapabilities.HasMMX()) cout << "MMX ";
  if (cpucapabilities.HasKNI()) cout << "KNI ";
  if (cpucapabilities.Has3dNow()) cout << "3dNow! ";
  if (cpucapabilities.HasMTRR()) cout << "MTRR ";
  if (cpucapabilities.HasCMOV()) cout << "CMOV ";
  if (cpucapabilities.HasFPU()) cout << "FPU ";
  cout << endl;
#endif

  try
    {
#if 1
      int c;

      while ((c=getopt(argc,argv,"f3rlcdv:wpuTSBbPHMQVW:XFYOCdmAa:h")) != EOF)
	{
	  switch (c)
	    {
	    case 'v':
	      {
#ifndef NDEBUG
		int level=atoi(optarg);
		if (level>7) level=7;
		if (level<0) level=0;
		switch(level)
		  {
		  case 7: vdec_options.Trace_SliceData = true;
		  case 6: vdec_options.Trace_DCTCoeff  = true;
		  case 5: vdec_options.Trace_MB   = true;
		  case 4: vdec_options.Trace_SlcH = true;
		  case 3: vdec_options.Trace_PicH = true;
		  case 2: vdec_options.Trace_GOPH = true;
		  case 1: vdec_options.Trace_SeqH = true;
		  case 0: break;
		  }
#else
		cerr << "Ignoring option -v#: Verbose output was not compiled in.\n";
#endif
	      }
	      break;
	      
	    case 'M':
	      if (!options.Postproc_ShowMBs)
		options.Postproc_ShowMBs = true;
	      else
		pp_mblks.SetBigMarks();
	      break;

	    case 'd':
	      pp_mblks.SetStaticMode(0);
	      break;

	    case 'l':
	      pp_mblks.SetStaticMode(255);
	      break;

	    case 'P':
	      vdec_options.DecodeP = vdec_options.DecodeB = false;
	      break;
	      
	    case 'B':
	      vdec_options.DecodeB = false;
	      break;
	      
	    case 'T':
	      imgtypes=true;
	      break;
	      
	    case 'Q':
	      qscales=true;
	      break;
	      
	    case 'V':
	      mvs=true;
	      break;

	    case 'C':
	      mvcol=true;
	      break;

	    case 'O':
	      mv_holdmode=true;
	      break;

	    case 'c': mvcolor=true; break;
	    case 'f': mvf=true; break;
	    case 'r': mvb=true; break;
	    case 'p': mvpf=true; break;
	    case 'b': mvbf=true; break;
	      
	    case 'A':
	      audio=true;
	      break;
	      
	    case '3':
	      ac3=true;
	      break;

	    case 'a':
	      ac3stream = atoi(optarg) + 0x80;
	      break;

	    case 'F':
	      if (!out_class) out_class = new Output_PPM;
	      break;

	    case 'Y':
	      if (!out_class) out_class = new Output_YUV;
	      break;

	    case 'h':
	      usage();
              exit(10);
	      break;
	    }
	}

      if (optind==argc)
	{ usage(); exit(5); }

      video_dec.SetOptions(vdec_options);

      ifstream str(argv[optind]);
      //system_dec.ForceStreamType(SystemDecoder_MPEG1::System);

      InitStreamSource(str);
      InitVideoDecoder1();
      if (audio) InitAudioDecoder();
      if (ac3)   InitAC3Decoder();

      if (options.Postproc_ShowMBs) InsertPostprocessor(pp_mblks);
      if (imgtypes) InsertPostprocessor(pp_imgtype);
      if (qscales)  InsertPostprocessor(pp_qscale);
      if (mvs)
	{
	  if (!mvf && !mvb) { mvf=mvb=true; }
	  pp_mv.SelectMVs(mvf,mvb);
	  pp_mv.ColoredVectors(mvcolor);
	  pp_mv.SelectFrametypes(mvpf,mvbf);
	  InsertPostprocessor(pp_mv);
	}
      if (mvcol)
	{
	  pp_mvcol.SetHoldMode(mv_holdmode);
	  //if (!mvf && !mvb) { mvf=mvb=true; }
	  //pp_mvcol2.SelectMVs(mvf,mvb);
	  pp_mvcol.SelectFrametypes(mvpf,mvbf);
	  InsertPostprocessor(pp_mvcol);
	}
      InitVideoDecoder2();

      //str.seekg(200000,ios::beg);

      if (!out_class)
	{
	  if (!default_x11server.AskDisplay())
	    { cerr << "Cannot open display.\n"; throw "no display"; }

	  out_class = new Output_Window;
	}

      ViewMPEG(audio,ac3);

      delete out_class;
#endif

#if 0
      ifstream str(argv[1]);
      ShowSystemPackets(str);
#endif

#if 0
      ifstream str(argv[1]);
      //ShowVideoAccessUnitPackets(str);
      //ShowVideoSyntax(str);
      DecodeVideo(str);
#endif

#if 0
      CheckMemAlloc();
#endif

#if 0
      if (!ProcessOptions(argc,argv, options))
	return 0;
#endif
    }
  catch (Excpt_Base& e)
    {
      MessageDisplay::Show(e.m_severity , e.m_text);
    }
  catch (...)
    {
      cerr << "error occurred... quitting...\n";
    }
}
